//#define DEBUG_LEVEL	0
//#define NVRIVA_ENABLE_MTRR

#define DRIVER_NAME "nVidia Riva Chipset driver"
#define DRIVER_REV  "0.1"

#if DEBUG_LEVEL > 1
#include "debug.h"
#endif

#include <kgi/system.h>
#include <kgi/module.h>
#include <IBM/vga_io.inc>

#include <nVidia/nvriva.h>
#include <nVidia/nv4ref.h>
#include <nVidia/nvreg.h>
#include <nVidia/riva_hw.h>

/* FIXME: MTRRs are a function of the CPU and cannot be assumed to exist */
//#include <asm/mtrr.h>

RIVA_HW_INST riva;
struct riva_regs old_state;
struct riva_regs newmode;
volatile unsigned *nvCONTROL;

struct riva_regs reg_template = 
{
	{ 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, /* ATTR */
	  0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
	  0x41, 0x01, 0x0F, 0x13, 0x00 },
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* CRT  */
	  0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE3, /* 0x10 */
	  0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 0x20 */
	  0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 0x30 */
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
          0x00,                                           /* 0x40 */
	},
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F, /* GRA  */
	  0xFF },
	{ 0x03, 0x01, 0x0F, 0x00, 0x0E},                  /* SEQ  */
	0xEB						  /* MISC */
};

#define SGS_VENDOR	0x12D2	/* RIVA 128 	*/
#define NV3_ID		0x0018  /* RIVA 128 	*/

#define NVIDIA_VENDOR	0x10DE  /* RIVA TNT 	*/
#define NV4_ID		0x0020  /* RIVA TNT 	*/

#define NV5_ID		0x0029	/* RIVA TNT2	*/

static void roundmode(kgi_modereq *gm,int roundx,int roundy);
static void nvriva_set_origin(struct kgi_display *dpy, kgi_uint x, uint y);
static void chipset_setup_dpy(struct kgi_display *dpy, struct kgi_mode *mode);

void save_state(struct kgi_display *dpy, struct riva_regs *regs);
void load_state(struct kgi_display *dpy, struct riva_regs *regs);

static struct kgi_chipset chipinfo =
{
	"nVidia",		/* Manufacturer		*/
	"NVx",			/* Model	(DETCTED) */
	GM_8BIT | GM_16BIT |	/* Available modes	*/
	GM_32BIT,
	0,			/* Flags		*/
	{ 2048, 2048 },		/* Maximum resolution	*/
	NULL,			/* Private data hook	*/
	0x0,			/* addressable RAM (DETECTED) */ 
	{ 0, 0 }		/* DCLK limits (DETECTED) */
};

static struct mem_region nvriva_control_mem =
{
	MEM_NULL,
	0,			/* linear control mem (DETECTED) */
	0,			/* mem size (DETECTED ??? ) */
	-1,			/* Decode all bits */
	"nVidia Riva linear control mem"
};

static struct mem_region nvriva_framebuffer_mem =
{
	MEM_NULL,
	0,			/* linear framebuffer (DETECTED) */
	0,			/* mem size (DETECTED) */
	-1,			/* Decode all bits */
	"nVidia Riva linear framebuffer"
};

#if 0
static struct io_region nvriva_vga_io =
{
	IO_NULL,
	0x3C0,			/* VGA IO Region Start */
	0x20,			/* 32 register (0x3C0-0x3DF) */
	0x3FF,			/* Only 10 address lines are decoded.. */
	"nVidia Riva VGA I/O region"
};    
#endif

const struct kgi_mmio_region nvriva_framebuffer_mmio = {
	NULL,                   	/* Next			*/
	"Riva MMIO framebuffer",	/* Name         	*/
	MMAP_FRAMEBUFFER,		/* Type (Linear)        */
	MMIO_RDWR | MMIO_SHARED,	/* Protection map       */

	0, 	   	                /* Total size of area (DETECTED) */
	0,           	         	/* Physical size       (DETECETED) */
	0,   	         	        /* bus address of the windows */
	0,                      	/* Logical adress       */

	NULL,                  		/* Set offset function  */
	NULL                    	/* private              */
};

void kgim_chipset_done(struct kgi_display *dpy)
{
	load_state(dpy, &old_state);
	mem_free_region( &nvriva_framebuffer_mem );
	mem_free_region( &nvriva_control_mem );
	kgi_free_small(KGIM_PRIVATE(dpy, chipset));
}
	 
static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == KGIGT_TEXT16)
		? mode->request.dpp.x : 8;

#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);
	if (mode->x.total % foo)
		mode->x.total += foo - (mode->x.total % foo);

	if (((mode->x.syncend / foo) & 0x3F) == ((mode->x.syncstart / foo) & 0x3F))
		mode->x.syncend -= foo;
	if (((mode->x.blankend / foo) & 0x7F) == ((mode->x.blankstart / foo) & 0x7F))
		mode->x.blankend -= foo;

	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
		mode->y.syncend--;
	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
		mode->y.blankend--;

#undef ADJUST
}

kgi_sint kgim_chipset_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
			     enum kgi_tm_cmd *cmd)
{
//	GET_PRIVATE;
	int bpp;

	DEBUG2("Checking mode... (%dx%d %d)\n", mode->request.virt.x,
		mode->request.virt.y, mode->request.graphtype);

	switch (mode->request.graphtype) {
	case KGIGT_8BIT:   bpp = 8; break;
	case KGIGT_15BIT:
	case KGIGT_16BIT:  bpp = 16; break;
	case KGIGT_32BIT:  bpp = 32; break;         
        default:
                DEBUG1("graphtype %i not supported", mode->request.graphtype);
                return -E(CHIPSET,FAILED);
        }
   
        #define XVIRTUAL (mode->request.virt.x)
        #define YVIRTUAL (mode->request.virt.y)
        #define XVISIBLE (mode->request.visible.x)
        #define YVISIBLE (mode->request.visible.y)
   
        switch (*cmd) {
		case CMD_PROPOSE:
		
		if (XVISIBLE <= 0 || YVISIBLE <= 0) {
			return -EINVAL;
		}
		
		if (XVIRTUAL == 0 && YVIRTUAL == 0)
			return -EINVAL;

//		if (XVIRTUAL == 0)
//			XVIRTUAL = fb_graph_len / YVIRTUAL);

//		if (YVIRTUAL == 0)
//			YVIRTUAL = fb_graph_len / (XVIRTUAL);

		mode->magnify.x = 1;
		mode->magnify.y = 1;
		roundmode(&mode->request, 8, 1);

//		while ((XVISIBLE * mode->magnify.x) < 640) {
//			mode->magnify.x++;
//		}

		while ((YVISIBLE * mode->magnify.y) < 350) {
			mode->magnify.y++;
		}

		/* Can't zoom more than this */
		if (mode->magnify.y > 2 || mode->magnify.x > 2) {
			return -EINVAL;
		}
         
		if ( (XVIRTUAL * YVIRTUAL * (bpp>>3) ) > chipinfo.ramsize) {

			if ( (XVISIBLE * YVISIBLE * (bpp>>3) ) > chipinfo.ramsize) {
				DEBUG1("Not enough memory to do "
			       		"virtual %ix%i at %i bpp",
			       		XVIRTUAL, YVIRTUAL, bpp);
				return -ENOMEM;
			}
			XVIRTUAL = XVISIBLE;
			YVIRTUAL = YVISIBLE;
		}

		mode->x.width = mode->request.visible.x * mode->magnify.x;
		mode->y.width = mode->request.visible.y * mode->magnify.y;
                return EOK;

        case CMD_LOWER:

                if (mode->dclk < chipinfo.dclk.min) {
                        DEBUG1("DCLK too low (%i)", mode->dclk);
                        return -E(CHIPSET,FAILED);
                }

                mode->dclk = (mode->dclk > chipinfo.dclk.max)
                        ? chipinfo.dclk.max : mode->dclk;

                adjust_timing(mode);
                return EOK;
      
        case CMD_RAISE:
		if (mode->dclk > chipinfo.dclk.max) {
                        DEBUG1("DCLK too high (%i)", mode->dclk);
                        return -E(CHIPSET,FAILED);
                }

                mode->dclk = (mode->dclk < chipinfo.dclk.min)
                        ? chipinfo.dclk.min : mode->dclk;
      
                adjust_timing(mode);
                return EOK;
      
        case CMD_CHECK:

                if (mode->lclk.mul != 1 && mode->lclk.mul != 3) {
                        DEBUG1("Invalid lclk multiplier %d", mode->lclk.mul);
                        return -E(CHIPSET,FAILED);
                }

		switch (mode->lclk.div) {
		case 1:
		case 2:
		case 4:
		case 8:
		case 16:
		case 32:
			break;
		default:
			DEBUG1("invalid CLK to LCLK divider %i", mode->lclk.div);
			return -E(CHIPSET, UNKNOWN);
		}

		mode->fb_size = mode->fsize =
			XVIRTUAL * YVIRTUAL * (bpp >> 3);

		mode->mem = NULL;
		mode->io  = NULL;

		mode->fb = (char *)nvriva_framebuffer_mem.baseptr;
		mode->fb_size = chipinfo.ramsize;

		mode->mmio = &nvriva_framebuffer_mmio;

                return EOK;
	default:
		return -EINVAL;
	}

        #undef XVIRTUAL
        #undef YVIRTUAL
        #undef XVISIBLE
        #undef YVISIBLE
}

kgi_sint kgim_chipset_command(struct kgi_graphic *state, uint cmd, uint iobuf)
{
	kgi_coord  *coord;
	
	
	switch (cmd) {
	case CHIP_SETVISFRAME:
		coord = (kgi_coord *)(state->io_buf);
		if (coord->x < 0 || coord->x > state->dev.mode.virt.x)
			return -E(CHIPSET,BOGUS);
		
		if (coord->y < 0 || coord->y > state->dev.mode.virt.y)
			return -E(CHIPSET,BOGUS);
		
		kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
		
		return 0;
	default:
		break;                                                    
	}
	return -E(CHIPSET, UNKNOWN);	
}

static int verify_nvriva(struct kgi_display *dpy)
{
	int i;
	struct pci_dev *dev = NULL;

	if ( (dev = pci_find_device(SGS_VENDOR, NV3_ID, dev)) )
	{
		riva.Architecture = 3;
	}
	else if ( (dev = pci_find_device(NVIDIA_VENDOR, NV4_ID, dev)) )
	{
		riva.Architecture = 4;
	}
	else if ( (dev = pci_find_device(NVIDIA_VENDOR, NV5_ID, dev)) )
	{
		riva.Architecture = 5;
	}
	else
	{
		NOTICE("Could't find a nVidia Riva card on the PCI/AGP bus");
		return -E(CHIPSET, FAILED);
	}

	nvriva_control_mem.base = dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK;
	nvriva_control_mem.size = 0x00800000; // (DETECT)
	
	nvriva_framebuffer_mem.base = dev->base_address[1] & PCI_BASE_ADDRESS_MEM_MASK;
	nvriva_framebuffer_mem.size = 0x01000000; // (DETECT)
	
	if ( mem_check_region( &nvriva_control_mem ) )
		return -E(CHIPSET, FAILED);
	
	if ( mem_check_region( &nvriva_framebuffer_mem ) )
		return -E(CHIPSET, FAILED);

	NOTICE("nVidia Riva Card detected.\nControl mem=0x%X\nFramebuffer mem=0x%X",
		nvriva_control_mem.base, nvriva_framebuffer_mem.base);
	return EOK;
}

kgi_sint kgim_chipset_init(struct kgi_card *card,struct kgi_display *dpy)
{
	GET_PRIVATE;

	if ( verify_nvriva(dpy) != EOK )
	{
		NOTICE("Failed to detect a card or resources was busy");
		return -E(CHIPSET, FAILED);
	}
	
	priv = kgi_malloc_small( sizeof(struct chipset_private) );
	KGIM_PRIVATE(dpy, chipset) = priv;

	mem_claim_region(&nvriva_control_mem);
	mem_claim_region(&nvriva_framebuffer_mem);
//	io_claim_region(&nvriva_vga_io);

	nvCONTROL = nvriva_control_mem.baseptr;

	riva.EnableIRQ = 0;
	riva.IO      = (io_in8(0x3CC) & 0x01) ? 0x3D0 : 0x3B0;
	riva.PRAMDAC = (unsigned *)((char *)nvCONTROL+0x00680000);
	riva.PFB     = (unsigned *)((char *)nvCONTROL+0x00100000);
	riva.PFIFO   = (unsigned *)((char *)nvCONTROL+0x00002000);
	riva.PGRAPH  = (unsigned *)((char *)nvCONTROL+0x00400000);
	riva.PEXTDEV = (unsigned *)((char *)nvCONTROL+0x00101000);
	riva.PTIMER  = (unsigned *)((char *)nvCONTROL+0x00009000);
	riva.PMC     = (unsigned *)((char *)nvCONTROL+0x00000000);
	riva.FIFO    = (unsigned *)((char *)nvCONTROL+0x00800000);

	switch (riva.Architecture)
	{
		case 3:
		riva.PRAMIN  = (unsigned *)((char *)nvCONTROL+0x00C00000);
		break;
		case 4:
		case 5:
		riva.PCRTC   = (unsigned *)((char *)nvCONTROL+0x00600000);
		riva.PRAMIN  = (unsigned *)((char *)nvCONTROL+0x00710000);
		break;
		
	}

	RivaGetConfig(&riva);

	chipinfo.ramsize = riva.RamAmountKBytes * 1024;
	NOTICE("%uKByte of RAM detected", riva.RamAmountKBytes );
	
	chipinfo.dclk.max = riva.MaxVClockFreqKHz * 1000;

	nvriva_framebuffer_mmio.size = nvriva_framebuffer_mmio.psize = chipinfo.ramsize;
	nvriva_framebuffer_mmio.pstart = nvriva_framebuffer_mem.base;

#ifdef NVRIVA_ENABLE_MTRR
	NOTICE1("Setting up MTRR");
	mtrr_add( nvriva_framebuffer_mem.base, chipinfo.ramsize, MTRR_TYPE_WRCOMB, 1);
	NOTICE1("MTRR setup done.");
#endif

	save_state(dpy, &old_state);
#if DEBUG_LEVEL > 1
	NOTICE("Original state:");
	dump_state(dpy, &old_state);
#endif

	return EOK;
}

/* Rounds up virt and visible fields to the next multiple of 'round'.
   Enlarges virt to be at least of size visible. */
   
static void roundmode(kgi_modereq *gm, int roundx, int roundy)
{
	gm->visible.x = ((gm->visible.x + roundx - 1) / roundx) * roundx;
	gm->visible.y = ((gm->visible.y + roundy - 1) / roundy) * roundy;
	gm->virt.x    = ((gm->virt.x + roundx - 1) / roundx) * roundx;
	gm->virt.y    = ((gm->virt.y + roundy - 1) / roundy) * roundy;

	if (gm->visible.x > gm->virt.x)
 	        gm->virt.x = gm->visible.x;
	if (gm->visible.y > gm->virt.y)
	        gm->virt.y = gm->visible.y;
}

void kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	IS_CRITICAL;
	
	GET_PRIVATE;
	int DCLKdiv = 8;
	int xtotal, xwidth, xblankstart, xblankend, xsyncstart, xsyncend,
	    ytotal, ywidth, yblankstart, yblankend, ysyncstart, ysyncend,
	    offset,
	    bpp;

	memcpy(&newmode, &reg_template, sizeof(struct riva_regs));

	priv->newmode = &newmode;

	DEBUG2("Setting mode... (%dx%d %d)\n", mode->request.virt.x,
		mode->request.virt.y, mode->request.graphtype);

	xtotal = (mode->x.total / DCLKdiv) - 5;
	xwidth = (mode->x.width / DCLKdiv) - 1;
	xblankstart = mode->x.blankstart / DCLKdiv;
	xblankend = mode->x.blankend / DCLKdiv;
	xsyncstart = mode->x.syncstart / DCLKdiv;
	xsyncend = mode->x.syncend / DCLKdiv;

	ytotal = mode->y.total - 2;
	ywidth = mode->y.width - 1;
	yblankstart = mode->y.blankstart - 1;
	yblankend = mode->y.blankend - 1;
	ysyncstart = mode->y.syncstart - 1;
	ysyncend = mode->y.syncend - 1;

	switch(mode->request.graphtype) {
	case KGIGT_8BIT:
		bpp = 8;
		break;
	case KGIGT_16BIT:
		bpp = 16;
		break;
	case KGIGT_32BIT:
		bpp = 32;
		break;
	default:
		INTERNAL_ERROR;
		return;
	}

	offset = (mode->request.virt.x/8) * (bpp/8);

	newmode.crtc[NV_PCRTC_HORIZ_TOTAL] =
		xtotal & 0xFF;

	newmode.crtc[NV_PCRTC_HORIZ_DISPLAY_END] =
		xwidth & 0xFF;

	newmode.crtc[NV_PCRTC_HORIZ_BLANK_START] =
		xblankstart & 0xFF;

	newmode.crtc[NV_PCRTC_HORIZ_BLANK_END] |=
		xblankend & 0x1F;

	newmode.crtc[NV_PCRTC_HORIZ_RETRACE_START] =
		xsyncstart & 0xFF;

	newmode.crtc[NV_PCRTC_HORIZ_RETRACE_END] =
		(xsyncend & 0x1F) |
		SetBitField(xblankend, 5:5, NV_PCRTC_HORIZ_RETRACE_END_HORIZ_BLANK_END_5);

//vertical
	newmode.crtc[NV_PCRTC_VERT_DISPLAY_END] =
		ywidth & 0xFF;

	newmode.crtc[NV_PCRTC_VERT_BLANK_START] =
		yblankstart & 0xFF;

	newmode.crtc[NV_PCRTC_VERT_RETRACE_START] =
		ysyncstart & 0xFF;

	newmode.crtc[NV_PCRTC_VERT_RETRACE_END] =
		ysyncend & 0xFF;

	newmode.crtc[NV_PCRTC_VERT_BLANK_END] =
		yblankend & 0xFF;

	newmode.crtc[NV_PCRTC_VERT_TOTAL] =
		ytotal & 0xFF;

/* other */

	newmode.crtc[NV_PCRTC_OVERFLOW] =
		SetBitField( ysyncstart,  9:9, NV_PCRTC_OVERFLOW_VERT_RETRACE_START_9) |
		SetBitField( ywidth,	  9:9, NV_PCRTC_OVERFLOW_VERT_DISPLAY_END_9) |
		SetBitField( ytotal,      9:9, NV_PCRTC_OVERFLOW_VERT_TOTAL_9) |
		SetBitField( 1,		  0:0, NV_PCRTC_OVERFLOW_LINE_COMPARE_8) |
		SetBitField( yblankstart, 8:8, NV_PCRTC_OVERFLOW_VERT_BLANK_START_8) |
		SetBitField( ysyncstart,  8:8, NV_PCRTC_OVERFLOW_VERT_RETRACE_START_8) |
		SetBitField( ywidth,      8:8, NV_PCRTC_OVERFLOW_VERT_DISPLAY_END_8) |
		SetBitField( ytotal,      8:8, NV_PCRTC_OVERFLOW_VERT_TOTAL_8);

	newmode.crtc[NV_PCRTC_OFFSET] =
		offset & 0xFF;

	newmode.crtc[NV_PCRTC_MAX_SCAN_LINE] =
		SetBitField( (mode->magnify.y == 2), 0:0, NV_PCRTC_MAX_SCAN_LINE_DOUBLE_SCAN) |
		SetBitField(1,             	     0:0, NV_PCRTC_MAX_SCAN_LINE_LINE_COMPARE_9) |
		SetBitField(yblankstart,	     9:9, NV_PCRTC_MAX_SCAN_LINE_VERT_BLANK_START_9) |
		SetBitField(0,			     4:0, NV_PCRTC_MAX_SCAN_LINE_MAX_SCAN_LINE);

        newmode.misc_output |=
		((mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC) |
		((mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC);

	/*	Ok.  We'll do the actual hardware programming now.
	*/

        kgim_clock_set_mode(dpy, mode);

	riva.CalcStateExt( &riva, &newmode.ext, bpp,
			   mode->request.virt.x,
			   mode->x.width,
			   xwidth,
			   xsyncstart,
			   xsyncend,
			   xtotal,
			   mode->request.virt.y,
			   ywidth,
			   ysyncstart,
			   ysyncend,
			   ytotal,
			   priv->clock / 1000); /* the clock is calculated in kgim_clock_set_mode */

	ENTER_CRITICAL;
	load_state(dpy, &newmode);
	kgim_monitor_set_mode(dpy, mode);
	kgim_ramdac_set_mode(dpy, mode);
	kgim_accel_set_mode(dpy, mode);
	LEAVE_CRITICAL;

	NOTICE2("setting up display...");

	chipset_setup_dpy(dpy, mode);

	NOTICE1("Mode set complete");
#if DEBUG_LEVEL > 1
	NOTICE("Newly programmed state:");
	dump_state(dpy, &newmode);
#endif
}

void save_state(struct kgi_display *dpy, struct riva_regs *regs)
{
	int i;

	io_out8(NV_CIO_SR_LOCK_INDEX, 0x3D4);
	io_out8(NV_CIO_SR_UNLOCK_RW_VALUE, 0x3D5);
	
	riva.UnloadStateExt(&riva, &regs->ext);
	
	regs->misc_output = io_in8(0x3CC);
	
	for (i = 0; i < NUM_CRT_REGS; i++) 
	{
		io_out8(i, 0x3D4);
		regs->crtc[i] = io_in8(0x3D5);
	}

	for (i = 0; i < NUM_ATC_REGS; i++) 
	{
		io_out8(i, 0x3C0);
		regs->attr[i] = io_in8(0x3C1);
	}

	for (i = 0; i < NUM_GRC_REGS; i++) 
	{
		io_out8(i, 0x3CE);
		regs->gra[i] = io_in8(0x3CF);
	}


	for (i = 0; i < NUM_SEQ_REGS; i++) 
	{
		io_out8(i, 0x3C4);
		regs->seq[i] = io_in8(0x3C5);
	}
}

void load_state(struct kgi_display *dpy, struct riva_regs *regs)
{
	int i;

	io_out8(0x11, 0x3D4);
	io_out8(0x00, 0x3D5);

	io_out8(NV_CIO_SR_LOCK_INDEX, 0x3D4);
	io_out8(NV_CIO_SR_UNLOCK_RW_VALUE, 0x3D5);
	
	riva.LoadStateExt(&riva, &regs->ext);
	
	io_out8(regs->misc_output, 0x3C2);
	
	for (i = 0; i < NUM_CRT_REGS; i++) 
	{
		if (i < 0x19)
		{
			io_out8(i, 0x3D4);
			io_out8(regs->crtc[i], 0x3D5);
		}
		else 
		{
			switch (i)
			{
				case 0x19:
				case 0x20:
				case 0x21:
				case 0x22:
				case 0x23:
				case 0x24:
				case 0x25:
				case 0x26:
				case 0x27:
//				case 0x28:
				case 0x29:
				case 0x2a:
				case 0x2b:
				case 0x2c:
				case 0x2d:
				case 0x2e:
				case 0x2f:
				case 0x30:
				case 0x31:
				case 0x32:
				case 0x33:
				case 0x34:
				case 0x35:
				case 0x36:
				case 0x37:
				case 0x38:
				case 0x39:
				case 0x3a:
				case 0x3b:
				case 0x3c:
				case 0x3d:
				case 0x3e:
				case 0x3f:
//				case 0x40:
				break;
				default:
				io_out8(i, 0x3D4);
				io_out8(regs->crtc[i], 0x3D5);
			}
		}
	}
	
	for (i = 0; i < NUM_ATC_REGS; i++) 
	{
		io_out8(i, 0x3C0);
		io_out8(regs->attr[i], 0x3C1);
	}
	
	for (i = 0; i < NUM_GRC_REGS; i++) 
	{
		io_out8(i, 0x3CE);
		io_out8(regs->gra[i], 0x3CF);
	}

	for (i = 0; i < NUM_SEQ_REGS; i++) 
	{
		io_out8(i, 0x3C4);
		io_out8(regs->seq[i], 0x3C5);
	}
}

static void nvriva_set_origin(struct kgi_display *dpy, kgi_uint x, uint y)
{
	IS_CRITICAL;

	ENTER_CRITICAL;

	riva.SetStartAddress(&riva, ((y * dpy->mode.request.virt.x) + x) * (newmode.ext.bpp/8) );

	LEAVE_CRITICAL;
}

static void chipset_setup_dpy(struct kgi_display *dpy, struct kgi_mode *mode)
{
	memcpy(&(dpy->mode), mode, sizeof(dpy->mode));
	
	dpy->set_font		= NULL;
	dpy->set_origin		= nvriva_set_origin;
	dpy->set_split		= NULL; //nvriva_set_split;
	dpy->set_frame		= NULL; //nvriva_set_frame;

	dpy->show_cursor	= NULL;
	dpy->hide_cursor	= NULL;
	dpy->undo_cursor	= NULL;

	dpy->show_pointer	= NULL;
	dpy->hide_pointer	= NULL;
	dpy->undo_pointer	= NULL;

	dpy->fb.size.x		= mode->request.virt.x;
	dpy->fb.size.y		= mode->request.virt.y;
	dpy->fb.pbuf		= NULL;
	dpy->fb.zbuf		= NULL;
	dpy->fb.ops		= NULL;
}

